/*-
 * Copyright (c) 2008 Andrey Simonenko
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "config.h"

#ifndef lint
static const char rcsid[] ATTR_UNUSED =
  "@(#)$Id: ipa_conv.c,v 1.2 2011/01/23 19:01:54 simon Exp $";
#endif /* !lint */

#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <regex.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <ipa_mod.h>

#include "queue.h"

#if (SIZEOF_LONG * CHAR_BIT) < 32
# error "'long' is less than 32 bits: too small"
#endif

#if IPA_AC_MOD_API_VERSION != 1 && IPA_AC_MOD_API_VERSION != 2
# error "This module supports only ipa_ac_mod API versions 1 and 2"
#endif

#if IPA_DB_MOD_API_VERSION != 1 && IPA_DB_MOD_API_VERSION != 2
# error "This module supports only ipa_db_mod API versions 1 and 2"
#endif

#if IPA_MEMFUNC_API_VERSION != 1
# error "This module supports only ipa_memfunc API version 1"
#endif

#ifdef WITH_PTHREAD
# define MOD_FLAGS IPA_MOD_FLAG_PTHREAD_SAFE
#else
# define MOD_FLAGS 0
#endif

#define MOD_NAME		"conv"
#define MOD_DB_CONF_PREFIX	"conv"
#if IPA_AC_MOD_API_VERSION == 1
# define MOD_AC_CONF_PREFIX	"conv_unused"
#else
# define MOD_AC_CONF_PREFIX	NULL
#endif

#define MARRAY_NAME(x)	MOD_NAME ":" #x
#define MZONE_NAME(x)	MOD_NAME ":" #x
#define MTYPE_NAME(x)	MOD_NAME ":m_" #x

/* Are defined at the end of this file. */
extern struct ipa_ac_mod IPA_AC_MOD_ENTRY(ipa_conv);
extern struct ipa_db_mod IPA_DB_MOD_ENTRY(ipa_conv);

static const ipa_suppfunc *suppfunc;    /* Exported support functions. */

static const ipa_memfunc *memfunc;      /* Exported memory functions. */

static ipa_mem_type *m_anon;		/* Anonymous memory allocations. */

static char	*(*mem_strdup)(const char *, ipa_mem_type *);
static void	*(*mem_malloc)(size_t, ipa_mem_type *);
static void	(*mem_free)(void *, ipa_mem_type *);

static char	conf_init_calls = 0;	/* Number of conf_init() calls. */

static signed char debug = -1;		/* conv:debug parameter. */
static signed char scale = -1;		/* conv:scale parameter. */
static signed char quiet = -1;		/* conv:quiet parameter. */

/*
 * These values should be sorted by most specific time interval
 * format.  This is important for the link_tint() function.
 */
#define TINT_TYPE_YMD		0	/* yyyy.mm.dd */
#define TINT_TYPE_MD		1	/* mm.dd */
#define TINT_TYPE_M		2	/* mm */
#define TINT_TYPE_WDAY		3	/* wday */
#define TINT_TYPE_INT		4	/* xx:xx-xx:xx */
#define TINT_TYPE_DEFAULT	5	/* Default settings. */
#define TINT_TYPE_UNKNOWN	6	/* Not yet determined. */

#define TINT_CONV_DECREASE	0	/* Statistics is decreased. */
#define TINT_CONV_ZERO		1	/* Statistics is zeroed. */
#define TINT_CONV_INCREASE	2	/* Statistics is increased. */

/*
 * Time interval descriptor.
 */
struct tint {
	SLIST_ENTRY(tint) link;		/* For list building.		*/
	uint32_t	sorig;		/* Original statistics.		*/
	uint32_t	sconv;		/* Converted statistics.	*/
	uint64_t	max_div_sconv;	/* 1 + UINT64_MAX / sconv.	*/
	unsigned int	sec1;		/* h1:m1:00			*/
	unsigned int	sec2;		/* h2:m2:00			*/
	unsigned char	type;		/* TINT_TYPE_xxx		*/
	unsigned char	conv;		/* TINT_CONV_xxx		*/
	uint8_t		mon;		/* Month.			*/
	uint8_t		day;		/* Month day or week day.	*/
	unsigned int	year;		/* Year.			*/
};

SLIST_HEAD(tint_list, tint);

static struct tint tint_nul;		/* Synthetic ..:..-00:00 interval. */

/*
 * Converting plan.
 */
struct plan {
	const struct tint *tint;	/* Current tint.		*/
	struct tint_list tint_list;	/* List of time intervals.	*/
	char		*name;		/* Name of plan.		*/
	char		*file;		/* File with plan.		*/
	int		year;		/* Current year.		*/
	int		mon;		/* Current month.		*/
	int		mday;		/* Current month day.		*/
	STAILQ_ENTRY(plan) link;	/* For list building.		*/
};

static STAILQ_HEAD(, plan) plans_list;	/* List of all plans. */

/*
 * rule { conv:dest } value, used field depends on stage of parsing.
 */
union dest {
	char		*name;		/* Destination rule name.	*/
	struct rule	*rule;		/* Pointer to destination rule.	*/
};

/*
 * Module's data for each rule.
 */
struct rule {
	union dest	dest;		/* rule { conv:dest }		*/
	struct plan	*plan;		/* rule { conv:plan }		*/
	uint64_t	cnt;		/* Current statistics.		*/
	uint64_t	sorig;		/* Not converted statistics.	*/
	uint64_t	sconv;		/* Converted statistics.	*/
	int		sconv_pos;	/* Non-zero if sconv >= 0.	*/
	char		sorig_pos;	/* Non-zero if sorig >= 0.	*/
	char		flags;		/* RULEF_xxx			*/
	const char	*name;		/* Name of this rule.		*/
};

#define RULEF_REFERENCED	0x01	/* Referenced by conv:dest. */
#define RULEF_DB_INITED		0x02	/* Initialized by db_init_statrule. */
#define RULEF_AC_INITED		0x04	/* Initialized by ac_init_statrule. */
#define RULEF_ACTIVE		0x08	/* Rule is active. */

static struct rule **rules_ptr;		/* Array of pointers to rules. */
static struct rule *currule;		/* Current rule. */
static unsigned int nrules;		/* Length of rules_ptr. */
static ipa_marray *rules_ptr_marray;	/* Marray for array of ptrs to rules. */

#define RULE(x) rules_ptr[x]		/* Pointer to rule by number. */

#define RULE_NSIZE	30
#define RULE_NALLOC	20

/* Pattern h1:m1-h2:m2 */
#define PAT_INT "\
^[[:digit:]]{1,2}:[[:digit:]]{1,2}-[[:digit:]]{1,2}:[[:digit:]]{1,2}$"

/* Pattern year.mon.day */
#define PAT_YMD "\
^[[:digit:]]+\\.([[:digit:]]|[[:alpha:]])+\\.[[:digit:]]+$"

/* Pattern mon.day. */
#define PAT_MD "\
^([[:digit:]]|[[:alpha:]])+\\.[[:digit:]]+$"

static regex_t	re_int;			/* Compiled TINT_INT */
static regex_t	re_ymd;			/* Compiled TINT_YMD */
static regex_t	re_md;			/* Compiled TINT_MD */

#define MONTHES_IN_YEAR		(12)
#define DAYS_IN_WEEK		(7)
#define HOURS_IN_DAY		(24)
#define SECONDS_IN_MINUTE	(60)
#define MINUTES_IN_HOUR		(60)
#define SECONDS_IN_HOUR		(MINUTES_IN_HOUR * SECONDS_IN_MINUTE)
#define SECONDS_IN_DAY		(HOURS_IN_DAY * SECONDS_IN_HOUR)

#define TINT_TYPE_MAXLEN	64

static const char wdays_short[DAYS_IN_WEEK] = {
	'S', 'M', 'T', 'W', 'H', 'F', 'A'
};

static const char *month_name[MONTHES_IN_YEAR + 1] = {
    NULL, "jan", "feb", "mar", "apr", "may", "jun",
    "jul", "aug", "sep", "oct", "nov", "dec"
};

/*
 * Last month day, do not distinguish February 28th and 29th,
 * since year is unknown in some cases.
 */
static int const last_mday_arr[MONTHES_IN_YEAR + 1] =
    /*   Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec */
    {  0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

struct conf_re {
	const char	*pat;		/* Regular expression pattern. */
	regex_t		*re;		/* Compiled regular expression. */
};

static struct conf_re conf_re_tbl[] = {
	{ PAT_INT,	&re_int		},
	{ PAT_YMD,	&re_ymd		},
	{ PAT_MD,	&re_md		}
};

#define CONF_RE_TBL_SIZE (sizeof(conf_re_tbl) / sizeof(conf_re_tbl[0]))

#define QUIETF_LOST_STAT	0x01	/* Check lost statistics on exit. */
#define QUIETF_NOT_CONVERTED	0x02	/* Check not converted statistics. */

#define QUIETF_ALL_FLAGS	(QUIETF_LOST_STAT|QUIETF_NOT_CONVERTED)

static void	logconf(const char *, ...) ATTR_FORMAT(printf, 1, 2);
static void	logconfx(const char *, ...) ATTR_FORMAT(printf, 1, 2);
static void	logmsgx(int, const char *, ...) ATTR_FORMAT(printf, 2, 3);

/*
 * A wrapper for imported logmsg() function,
 * but error code is always zero here.
 */
static void
logmsgx(int priority, const char *format, ...)
{
	va_list ap;

	va_start(ap, format);
	suppfunc->logmsg(MOD_NAME, priority, 0, format, ap);
	va_end(ap);
}

/*
 * A wrapper for imported logconferr() function.
 */
static void
logconf(const char *format, ...)
{
	va_list ap;
	int errno_save;

	errno_save = errno;
	va_start(ap, format);
	suppfunc->logconferr(MOD_NAME, errno_save, format, ap);
	va_end(ap);
}

/*
 * A wrapper for imported logconferr() function,
 * but error code is always zero here.
 */
static void
logconfx(const char *format, ...)
{
	va_list ap;

	va_start(ap, format);
	suppfunc->logconferr(MOD_NAME, 0, format, ap);
	va_end(ap);
}

/*
 * A wrapper for imported print_param_args() function.
 */
static void
print_param_args(const char *format, ...)
{
	va_list ap;

	va_start(ap, format);
	suppfunc->print_param_args(format, ap);
	va_end(ap);
}

/*
 * Simplified version of regexec(3).
 */
static int
regexec_simple(const regex_t *re, const char *str)
{
	return (regexec(re, str, (size_t)0, (regmatch_t *)NULL, 0));
}

static int
build_conf_re(void)
{
	char buf[128];
	const struct conf_re *cre;
	unsigned int i;
	int error;

	for (i = 0, cre = conf_re_tbl; i < CONF_RE_TBL_SIZE; ++cre, ++i) {
		error = regcomp(cre->re, cre->pat, REG_EXTENDED|REG_NOSUB);
		if (error != 0) {
			regerror(error, (regex_t *)NULL, buf, sizeof(buf));
			logconfx("build_conf_re: regcomp(\"%s\"): %s",
			    cre->pat, buf);
			return (-1);
		}
	}
	return (0);
}

static void
free_conf_re(void)
{
	const struct conf_re *cre;
	unsigned int i;

	for (i = 0, cre = conf_re_tbl; i < CONF_RE_TBL_SIZE; ++cre, ++i)
		regfree(cre->re);
}

/*
 * Return greatest common divisor.
 */
static uint32_t
gcd(uint32_t a, uint32_t b)
{
	uint32_t c;

	for (; b != 0;) {
		c = a;
		a = b;
		b = c % b;
	}
	return (a);
}

/*
 * Check whether a given string has at least one '\"'.
 */
static int
is_string(const char *s)
{
	for (; *s != '\0'; ++s)
		if (*s == '\"')
			return (1);
	return (0);
}

static int
alloc_rule(unsigned int no, const char *name)
{
	struct rule *rule;

	rule = mem_malloc(sizeof(*rule), m_anon);
	if (rule == NULL) {
		logconfx("mem_malloc failed");
		return (-1);
	}
	rule->name = name;
	rule->dest.name = NULL;
	rule->plan = NULL;
	rule->sorig = rule->sconv = 0;
	rule->flags = 0;

	if (memfunc->marray_alloc(rules_ptr_marray, &no, 1) < 0) {
		logconfx("marray_alloc failed");
		return (-1);
	}
	RULE(no) = currule = rule;

	return (0);
}

/*
 * Convert time interval to string.  buf_size must be
 * greater than or equal to 2.
 */
static void
tint_to_str(const struct tint *tint, char *buf, size_t buf_size)
{
	unsigned int h1, m1, h2, m2;
	int n;
	char buf1[sizeof(" xx:xx-xx:xx")];

	switch (tint->type) {
	case TINT_TYPE_YMD:
		n = snprintf(buf, buf_size, "%u.%s.%02u",
		    tint->year, month_name[tint->mon], tint->day);
		break;
	case TINT_TYPE_MD:
		n = snprintf(buf, buf_size, "%s.%02u",
		    month_name[tint->mon], tint->day);
		break;
	case TINT_TYPE_M:
		n = snprintf(buf, buf_size, month_name[tint->mon]);
		break;
	case TINT_TYPE_WDAY:
		n = 1;
		buf[0] = wdays_short[tint->day];
		buf[1] = '\0';
		break;
	case TINT_TYPE_INT:
		n = 0;
		break;
	default: /* TINT_TYPE_DEFAULT */
		buf[0] = '*';
		buf[1] = '\0';
		return;
	}

	if (n >= buf_size)
		goto failed;

	h1 = tint->sec1 / SECONDS_IN_HOUR;
	m1 = (tint->sec1 % SECONDS_IN_HOUR) / SECONDS_IN_MINUTE;
	h2 = tint->sec2 / SECONDS_IN_HOUR;
	m2 = (tint->sec2 % SECONDS_IN_HOUR) / SECONDS_IN_MINUTE;
	if (tint->type == TINT_TYPE_INT ||
	    !(tint->sec1 == 0 && tint->sec2 == SECONDS_IN_DAY)) {
		if (snprintf(buf1, sizeof(buf1), " %02u:%02u-%02u:%02u",
		    h1, m1, h2, m2) != sizeof(buf1) - 1)
			goto failed;
	} else
		buf1[0] = '\0';

	if (buf_size < n + sizeof(buf1))
		goto failed;
	if (n == 0)
		strcpy(buf, buf1 + 1);
	else if (buf1[0] != '\0') {
		buf1[0] = '/';
		strcpy(buf + n, buf1);
	}
	return;

failed:
	buf[0] = '?';
	buf[1] = '\0';
}

/*
 * A wrapper for imported print_param_name() function.
 */
static void
print_param_name(const char *name)
{
	suppfunc->print_param_name(MOD_DB_CONF_PREFIX, name);
}

/*
 * A wrapper for imported print_param_end() function.
 */
static void
print_param_end(void)
{
	suppfunc->print_param_end();
}

static struct plan *
plan_by_name(const char *name)
{
	struct plan *plan;

	STAILQ_FOREACH(plan, &plans_list, link)
		if (strcmp(plan->name, name) == 0)
			break;
	return (plan);
}

/*
 * Skip all space and tab characters starting from s.
 * The '#' character marks the end of line.
 */
static char *
skip_spaces(char *s)
{
	for (; *s != '\0'; ++s)
		if (*s != ' ' && *s != '\t') {
			if (*s == '#')
				*s = '\0';
			break;
		}
	return (s);
}

/*
 * Skip all characters that belong to data and following space and
 * tab characters.  The '#' character marks the end of line.
 */
static char *
skip_data_spaces(char *s)
{
	for (; *s != '\0'; ++s) {
		if (*s == ' ' || *s == '\t') {
			*s = '\0';
			s = skip_spaces(s + 1);
			break;
		}
		if (*s == '#') {
			*s = '\0';
			break;
		}
	}
	return (s);
}

/*
 * Convert string to 'unsigned int'.
 * The first character in a string should be checked
 * for '-' before calling this function (remember that
 * strtoul() and strtoull() work with negative values).
 */
static int
strto_u_int(unsigned int *result, const char *nptr)
{
	char *endptr;
	unsigned long val_ul;

	if (isdigit((unsigned char)*nptr) == 0) {
		logconfx("\"%s\" is not an unsigned integer", nptr);
		return (-1);
	}
	errno = 0;
	val_ul = strtoul(nptr, &endptr, 10);
	if (errno != 0) {
		logconf("strtoul");
		return (-1);
	}
#if ULONG_MAX > UINT_MAX
	if (val_ul > UINT_MAX) {
		logconfx("too big value for 'unsigned int'");
		return (-1);
	}
#endif
	if (nptr == endptr || *endptr != '\0') {
		logconfx("wrong number");
		return (-1);
	}
	*result = (unsigned int)val_ul;
	return (0);
}

/*
 * Convert string to uint32_t.
 */
static int
strto_uint32(uint32_t *result, const char *nptr)
{
	char *endptr;
	unsigned long val;

	if (isdigit((unsigned char)*nptr) == 0) {
		logconfx("\"%s\" is not an unsigned integer", nptr);
		return (-1);
	}
	errno = 0;
	val = strtoul(nptr, &endptr, 10);
	if (errno != 0) {
		logconf("strtoul");
		return (-1);
	}
#if (SIZEOF_LONG * CHAR_BIT) > 32
	if (val > UINT32_MAX) {
		logconfx("too big value for 'uint32_t' type");
		return (-1);
	}
#endif
	if (nptr == endptr || *endptr != '\0') {
		logconfx("wrong number");
		return (-1);
	}
	*result = (uint32_t)val;
	return (0);
}

static int
parse_tint_month(const char *s, uint8_t *mon)
{
	unsigned int i;

	if (isdigit((unsigned char)*s)) {
		if (strto_u_int(&i, s) < 0 || i == 0 || i > MONTHES_IN_YEAR) {
			logconfx("wrong month number %u", i);
			return (-1);
		}
	} else {
		for (i = 1; i <= MONTHES_IN_YEAR; ++i)
			if (strcasecmp(s, month_name[i]) == 0)
				break;
		if (i > MONTHES_IN_YEAR) {
			logconfx("wrong month name \"%s\"", s);
			return (-1);
		}
	}
	*mon = (uint8_t)i;
	return (0);
}

static int
parse_tint_day(const char *s, uint8_t mon, uint8_t *day)
{
	unsigned int v;

	if (strto_u_int(&v, s) < 0 || v == 0 || v > last_mday_arr[mon]) {
		logconfx("wrong month day number");
		return (-1);
	}
	*day = (uint8_t)v;
	return (0);
}

/*
 * Parse time interval.
 */
static int
parse_tint(char *p, struct tint *tint)
{
	char *t, *pmon, *pday;
	unsigned int sec1, sec2;
	uint8_t mon;
	unsigned char type;

	type = TINT_TYPE_UNKNOWN;
	t = strchr(p, '/');
	if (t == NULL) {
		t = strchr(p, ':');
		if (t != NULL) {
			/* Range of hours and minutes. */
			type = TINT_TYPE_INT;
			t = p;
		} else if (*p == '*' && *(p + 1) == '\0') {
			/* The default settings. */
			type = TINT_TYPE_DEFAULT;
		}
	}

	if (t != NULL) {
		unsigned short h1, h2, m1, m2;

		if (*t == '/') {
			*t++ = '\0';
			if (*t == '\0')
				return (-1);
		}
		if (regexec_simple(&re_int, t) != 0) {
			logconfx("incorrect range of hours and minutes");
			return (-1);
		}
		if (sscanf(t, "%hu:%hu-%hu:%hu", &h1, &m1, &h2, &m2) != 4) {
			logconfx("parse_tint: sscanf(\"%s\") failed", t);
			return (-1);
		}
		if ((h1 > 23 || h2 > 23 || m1 > 59 || m2 > 59) &&
		    !(h2 == HOURS_IN_DAY && m2 == 0)) {
			logconfx("wrong value of hours or minutes in time "
			    "interval");
			return (-1);
		}
		if (h1 == h2 && m1 == m2) {
			logconfx("zero seconds time intervals are not "
			    "allowed");
			return (-1);
		}
		sec1 = h1 * SECONDS_IN_HOUR + m1 * SECONDS_IN_MINUTE;
		sec2 = h2 * SECONDS_IN_HOUR + m2 * SECONDS_IN_MINUTE;
		if (sec1 > sec2) {
			logconfx("incorrect range of hours and minutes");
			return (-1);
		}
	} else {
		sec1 = 0;
		sec2 = SECONDS_IN_DAY;
	}

	if (type != TINT_TYPE_UNKNOWN) {
		;
	} else if (regexec_simple(&re_ymd, p) == 0) {
		/* Year, month, day. */
		pmon = strchr(p, '.');
		*pmon++ = '\0';
		pday = strchr(pmon, '.');
		*pday++ = '\0';
		if (strto_u_int(&tint->year, p) < 0) {
			logconfx("wrong year number");
			return (-1);
		}
		if (parse_tint_month(pmon, &mon) < 0)
			return (-1);
		if (parse_tint_day(pday, mon, &tint->day) < 0)
			return (-1);
		tint->mon = mon;
		type = TINT_TYPE_YMD;
	} else if (regexec_simple(&re_md, p) == 0) {
		/* Month, day. */
		pday = strchr(p, '.');
		*pday++ = '\0';
		if (parse_tint_month(p, &mon) < 0)
			return (-1);
		if (parse_tint_day(pday, mon, &tint->day) < 0)
			return (-1);
		tint->mon = mon;
		type = TINT_TYPE_MD;
	} else if (isalpha((unsigned char)*p) && *(p + 1) == '\0') {
		/* Week day. */
		p = strchr(wdays_short, *p);
		if (p == NULL) {
			logconfx("wrong week day");
			return (-1);
		}
		tint->day = (uint8_t)(p - wdays_short);
		type = TINT_TYPE_WDAY;
	} else {
		/* Month. */
		if (t != NULL)
			return (-1);
		if (parse_tint_month(p, &mon) < 0)
			return (-1);
		tint->mon = mon;
		type = TINT_TYPE_M;
	}

	tint->sec1 = sec1;
	tint->sec2 = sec2;
	tint->type = type;

	return (0);
}

/*
 * Check whether the new time interval 'tint0' overlaps with
 * another time interval in the 'list' and add it to the 'list'.
 * All time intervals must be sorted by TINT_TYPE_xxx values.
 */
static int
link_tint(struct tint_list *list, struct tint *tint0)
{
	struct tint *tint, *tint_prev;

	tint_prev = NULL;
	SLIST_FOREACH(tint, list, link) {
		if (tint->type > tint0->type) {
			/* Should be inserted after prev. interval. */
			break;
		}
		if (tint->type < tint0->type) {
			/* Should be inserted somewhere after this one. */
		} else {
			/* Check overlap. */
			switch (tint->type) {
			case TINT_TYPE_YMD:
				if (tint->year != tint0->year ||
				    tint->mon != tint0->mon ||
				    tint->day != tint0->day)
					continue;
				break;
			case TINT_TYPE_MD:
				if (tint->mon != tint0->mon ||
				    tint->day != tint0->day)
					continue;
				break;
			case TINT_TYPE_M:
				if (tint->mon != tint0->mon)
					continue;
				break;
			case TINT_TYPE_WDAY:
				if (tint->day != tint0->day)
					continue;
				break;
			case TINT_TYPE_DEFAULT:
				logconfx("duplicated default time interval");
				return (-1);
			}
			if (!(tint0->sec2 <= tint->sec1 ||
			    tint0->sec1 >= tint->sec2)) {
				char buf1[TINT_TYPE_MAXLEN];
				char buf2[TINT_TYPE_MAXLEN];

				tint_to_str(tint, buf1, sizeof(buf1));
				tint_to_str(tint0, buf2, sizeof(buf2));
				logconfx("time interval '%s' overlaps time "
				    "interval '%s'", buf1, buf2);
				return (-1);
			}
			if (tint0->sec2 <= tint->sec1) {
				/* Should be inserted after prev. interval. */
				break;
			}
		}
		tint_prev = tint;
	}
	if (tint_prev != NULL)
		SLIST_INSERT_AFTER(tint_prev, tint0, link);
	else
		SLIST_INSERT_HEAD(list, tint0, link);
	return (0);
}

/*
 * Parse "conv:load" parameter.
 */
static int
parse_load(void *arg)
{
	const char *file, *name;
	struct plan *plan;
	struct tint *tint;
	FILE *fp;
	char *p1, *p2, *p3;
	unsigned int lineno;
	uint32_t gcd_value;
	char buf[257], has_default;

	name = *(char **)arg;
	p1 = strchr(name, ' ');
	*p1++ = '\0';

	if (is_string(name)) {
		logconfx("plan name cannot be a string");
		return (-1);
	}

	/* Validate arguments. */
	if (plan_by_name(name) != NULL) {
		logconfx("duplicated plan name \"%s\"", name);
		return (-1);
	}
	if (*p1++ != '\"') {
		logconfx("file name should be given as a string");
		return (-1);
	}
	if (*p1 != '/') {
		logconfx("file name should be given with absolute path");
		return (-1);
	}

	/* Fetch file name from string argument. */
	file = p1;
	for (; *p1 != '\"';)
		if (*p1 == '\\') {
			switch (*(p1 + 1)) {
			case 't':
				*p1 = '\t';
				break;
			case 'n':
				*p1 = '\n';
				break;
			default:
				*p1 = *(p1 + 1);
			}
			p1 += 2;
		} else
			++p1;
	*p1 = '\0';

	/* Create new plan structure. */
	plan = mem_malloc(sizeof(*plan), m_anon);
	if (plan == NULL) {
		logconfx("mem_malloc failed");
		return (-1);
	}
	if ((plan->name = mem_strdup(name, m_anon)) == NULL ||
	    (plan->file = mem_strdup(file, m_anon)) == NULL) {
		logconfx("mem_strdup failed");
		return (-1);
	}
	plan->tint = &tint_nul;
	plan->mon = MONTHES_IN_YEAR;
	STAILQ_INSERT_HEAD(&plans_list, plan, link);
	SLIST_INIT(&plan->tint_list);

	/* Parse the content of the plan file. */
	fp = fopen(file, "r");
	if (fp == NULL) {
		logconf("fopen(%s, \"r\")", file);
		return (-1);
	}

	has_default = 0;
	lineno = 0;
	while (fgets(buf, (int)sizeof(buf), fp) != NULL) {
		/* Check line length. */
		++lineno;
		p1 = strchr(buf, '\0');
		if (p1 == &buf[sizeof(buf) - 1]) {
			logconfx("too long line, max line length is %u\n",
			    (unsigned int)(sizeof(buf) - 2));
			goto failed;
		}

		/* Remove newline character. */
		if (*--p1 == '\n')
			*p1 = '\0';

		p1 = skip_spaces(buf);
		if (*p1 == '\0') {
			/* Comment or empty line. */
			continue;
		}

		p2 = skip_data_spaces(p1);
		if (*p2 == '\0') {
			logconfx("second field is absent");
			goto failed;
		}

		p3 = skip_data_spaces(p2);
		if (*p3 == '\0') {
			logconfx("third field is absent");
			goto failed;
		}

		if (*skip_data_spaces(p3) != '\0') {
			logconfx("too many fields in one line");
			goto failed;
		}

		/* Parse first field. */
		tint = mem_malloc(sizeof(*tint), m_anon);
		if (tint == NULL) {
			logconfx("mem_malloc failed");
			return (-1);
		}
		if (parse_tint(p1, tint) < 0) {
			logconfx("wrong time interval");
			goto failed;
		}

		/* Parse second and third fields. */
		if (strto_uint32(&tint->sorig, p2) < 0) {
			logconfx("second field should be unsigned integer");
			goto failed;
		}
		if (tint->sorig == 0) {
			logconfx("second field should be greater than zero");
			goto failed;
		}
		if (strto_uint32(&tint->sconv, p3) < 0) {
			logconfx("third field should be unsigned integer");
			goto failed;
		}
		if (tint->sconv == 0)
			tint->conv = TINT_CONV_ZERO;
		else if (tint->sorig >= tint->sconv)
			tint->conv = TINT_CONV_DECREASE;
		else
			tint->conv = TINT_CONV_INCREASE;
		if (scale != 0) {
			gcd_value = gcd(tint->sorig, tint->sconv);
			if (gcd_value != 1) {
				tint->sorig /= gcd_value;
				tint->sconv /= gcd_value;
			}
		}
		tint->max_div_sconv = 1 + UINT64_MAX / tint->sconv;
		if (link_tint(&plan->tint_list, tint) < 0)
			goto failed;
		if (tint->type == TINT_TYPE_DEFAULT)
			has_default = 1;
	}

	if (ferror(fp) != 0) {
		logconf("fgets(%s)", file);
		goto failed;
	}
	if (fclose(fp) != 0) {
		logconf("fclose(%s)", file);
		return (-1);
	}

	if (!has_default) {
		logconfx("plan \"%s\" does not have default time interval",
		    name);
		return (-1);
	}
	return (0);

failed:
	logconfx("error occurred at line %u in file %s", lineno, file);
	if (fclose(fp) != 0)
		logconf("fclose(%s)", file);
	return (-1);
}

/*
 * Parse "conv:dest" parameter.
 */
static int
parse_dest(void *arg)
{
	struct rule *rule;
	char *name;

	rule = currule;
	if (rule->dest.name != NULL) {
		logconfx("this parameter is duplicated");
		return (-1);
	}

	name = *(char **)arg;
	if (is_string(name)) {
		logconfx("rule name cannot be a string");
		return (-1);
	}

	if (strcmp(rule->name, name) == 0) {
		logconfx("destination rule should be some another rule");
		return (-1);
	}

	name = mem_strdup(name, m_anon);
	if (name == NULL) {
		logconfx("mem_strdup failed");
		return (-1);
	}
	rule->dest.name = name;

	return (0);
}

/*
 * Parse "conv:plan" parameter.
 */
static int
parse_plan(void *arg)
{
	const char *name;
	struct rule *rule;
	struct plan *plan;

	rule = currule;
	if (rule->plan != NULL) {
		logconfx("this parameter is duplicated");
		return (-1);
	}

	name = *(char **)arg;
	plan = plan_by_name(name);
	if (plan == NULL) {
		logconfx("cannot find plan with name \"%s\"", name);
		return (-1);
	}
	rule->plan = plan;
	return (0);
}

/*
 * Parse "conv:debug" parameter.
 */
static int
parse_debug(void *arg)
{
	uint32_t val;

	val = (*(uint32_t *)arg);
	if (val > 1) {
		logconfx("too big debug level, max debug level is 1");
		return (-1);
	}
	debug = (signed char)val;
	return (0);
}

/*
 * Parse "conv:scale" parameter.
 */
static int
parse_scale(void *arg)
{
	if (!STAILQ_EMPTY(&plans_list)) {
		logconfx("this parameter should be used before conv:load "
		    "parameters");
		return (-1);
	}
	scale = *(int *)arg;
	return (0);
}

/*
 * Parse "conv:quiet" parameter.
 */
static int
parse_quiet(void *arg)
{
	uint32_t val;

	val = *(uint32_t *)arg;
	if (val > QUIETF_ALL_FLAGS) {
		logconfx("such value is not allowed");
		return (-1);
	}
	quiet = (signed char)val;
	return (0);
}

static int
ac_conf_init(void)
{
	++conf_init_calls;
	suppfunc = IPA_AC_MOD_ENTRY(ipa_conv).suppfunc;
	return (0);
}

static int
db_conf_init(void)
{
	++conf_init_calls;

	suppfunc = IPA_DB_MOD_ENTRY(ipa_conv).suppfunc;

	memfunc = IPA_DB_MOD_ENTRY(ipa_conv).memfunc;
	if (memfunc->api_ver != IPA_MEMFUNC_API_VERSION) {
		logconfx("module understands memfunc API version %u, "
		    "exported memfunc API version is %u",
		    IPA_MEMFUNC_API_VERSION, memfunc->api_ver);
		return (-1);
	}

	m_anon = memfunc->mem_type_new(MTYPE_NAME(anon), "Anonymous memory", 0);
	if (m_anon == NULL) {
		logconfx("mem_type_new failed");
		return (-1);
	}

	rules_ptr_marray = memfunc->marray_init(MARRAY_NAME(rules_ptr),
	    "Pointers to rules", 0, (void *)&rules_ptr,
	    sizeof(struct rule *), RULE_NSIZE, RULE_NALLOC);
	if (rules_ptr_marray == NULL) {
		logconfx("marray_init failed");
		return (-1);
	}

	mem_malloc = memfunc->mem_malloc;
	mem_strdup = memfunc->mem_strdup;
	mem_free = memfunc->mem_free;

	if (build_conf_re() < 0)
		return (-1);

	tint_nul.sec2 = 0;

	STAILQ_INIT(&plans_list);

	return (0);
}

static int
conf_deinit(void)
{
	static char called = 0;

	struct rule *rule1, *rule2;
	unsigned int i, j;

	if (called)
		return (0);
	called = 1;

	if (conf_init_calls != 2) {
		logconfx("this module had to be loaded simultaneously "
		    "as accounting and database module");
		logconfx("or your system loaded this module two times "
		    "via dlopen(3) or similar interface");
		return (-1);
	}

	free_conf_re();

	/* Find all destination rules. */
	for (i = 0; i < nrules; ++i) {
		rule1 = RULE(i);
		if (rule1->dest.name == NULL) {
			rule1->dest.rule = NULL;
			continue;
		}
		for (j = 0; j < nrules; ++j) {
			rule2 = RULE(j);
			if (strcmp(rule1->dest.name, rule2->name) == 0) {
				mem_free(rule1->dest.name, m_anon);
				rule1->dest.rule = rule2;
				rule2->flags |= RULEF_REFERENCED;
				break;
			}
		}
		if (j == nrules) {
			logconfx("rule %s: destination rule %s does not exist",
			    rule1->name, rule1->dest.name);
			return (-1);
		}
	}

	return (0);
}

static void
dump_plan_list(void)
{
	const struct plan *plan;
	const struct tint *tint;
	char buf[TINT_TYPE_MAXLEN];

	/* yyyy.mmm.dd/xx:xx-xx:xx -- longest line if yyyy is 4 digits. */
	logmsgx(IPA_LOG_INFO, "List of loaded plans (values are %sscaled):",
	    scale ? "" : "not ");
	STAILQ_FOREACH(plan, &plans_list, link) {
		logmsgx(IPA_LOG_INFO, " Plan %s, file %s",
		    plan->name, plan->file);
		SLIST_FOREACH(tint, &plan->tint_list, link) {
			tint_to_str(tint, buf, sizeof(buf));
			logmsgx(IPA_LOG_INFO, "  %-23s %10"PRIu32" %10"PRIu32,
			    buf, tint->sorig, tint->sconv);
		}
	}
}

static int
db_conf_mimic_real(void)
{
	if (debug < 0)
		debug = 0;
	if (scale < 0)
		scale = 1;
	if (quiet < 0)
		quiet = QUIETF_ALL_FLAGS;
	return (0);
}

static int
ac_db_init(void)
{
	static char called = 0;

	struct rule *rule;
	unsigned int i, n;

	if (++called != 2)
		return (0);

	db_conf_mimic_real();

	for (i = n = 0; i < nrules; ++i) {
		rule = RULE(i);
		if (rule->flags & RULEF_REFERENCED) {
			if (!(rule->flags & RULEF_AC_INITED)) {
				logmsgx(IPA_LOG_ERR, "rule %s: this rule is "
				    "referenced by some conv:dest parameter, "
				    "but it does not use module's accounting "
				    "system", rule->name);
				return (-1);
			}
		}
		if (rule->dest.rule != NULL) {
			if (!(rule->flags & RULEF_DB_INITED)) {
				logmsgx(IPA_LOG_ERR, "rule %s: this rule has "
				    "the conv:dest parameter, but it does not "
				    "use module's database", rule->name);
				return (-1);
			}
		}
		if (!(rule->flags & (RULEF_AC_INITED|RULEF_DB_INITED))) {
			/* This rule does not use this module. */
			mem_free(rule, m_anon);
			RULE(i) = NULL;
			memfunc->marray_free(rules_ptr_marray, i);
		} else
			n = i;
	}

	nrules = n + 1;

	if (debug)
		dump_plan_list();

	return (0);
}

/* ARGSUSED */
static void
ac_conf_show(unsigned int sect_id ATTR_UNUSED, unsigned int no ATTR_UNUSED)
{
}

static void
db_conf_show(unsigned int sect_id, unsigned int no)
{
	const struct plan *plan;
	const struct rule *rule;

	switch (sect_id) {
	case IPA_CONF_SECT_ROOT:
		if (debug >= 0) {
			print_param_name("debug");
			print_param_args("%d", debug);
			print_param_end();
		}
		if (scale >= 0) {
			print_param_name("scale");
			suppfunc->print_boolean(scale);
			print_param_end();
		}
		if (quiet >= 0) {
			print_param_name("quiet");
			print_param_args("%d", quiet);
			print_param_end();
		}
		STAILQ_FOREACH(plan, &plans_list, link) {
			print_param_name("load");
			print_param_args(plan->name);
			suppfunc->print_space();
			suppfunc->print_string(plan->file);
			print_param_end();
		}
		break;
	case IPA_CONF_SECT_RULE:
		rule = RULE(no);
		if (rule->dest.rule != NULL) {
			print_param_name("dest");
			print_param_args(rule->dest.rule->name);
			print_param_end();
			print_param_name("plan");
			print_param_args(rule->plan->name);
			print_param_end();
		}
		break;
	}
}

/* ARGSUSED */
static int
ac_conf_event(unsigned int event ATTR_UNUSED, unsigned int no ATTR_UNUSED,
    const void *arg ATTR_UNUSED)
{
	return (0);
}

static int
db_conf_event(unsigned int event, unsigned int no, const void *arg)
{
	switch (event) {
	case IPA_CONF_EVENT_RULE_BEGIN:
		nrules = no + 1;
		if (alloc_rule(no, arg) < 0)
			return (-1);
		break;
	case IPA_CONF_EVENT_RULE_END:
		if ((currule->dest.name == NULL && currule->plan != NULL) ||
		    (currule->dest.name != NULL && currule->plan == NULL)) {
			logconfx("rule %s: rule should have both conv:dest "
			    "and conv:plan parameters", currule->name);
			return (-1);
		}
		break;
	}
	return (0);
}

static int
ac_conf_mimic_real(void)
{
	return (0);
}

static int
ac_pre_init(void)
{
	return (0);
}

static int
ac_init_statrule(unsigned int ruleno, const char *rule_name)
{
	struct rule *rule;

	rule = RULE(ruleno);
	if (!(rule->flags & RULEF_REFERENCED)) {
		logmsgx(IPA_LOG_ERR, "rule %s: this rule uses module's "
		    "accounting system, but it is not referenced by any "
		    "conv:dest parameter", rule_name);
		return (-1);
	}
	rule->flags |= RULEF_AC_INITED|RULEF_ACTIVE;
	return (0);
}

static int
ac_deinit_rule(unsigned int ruleno)
{
	if (!(quiet & QUIETF_LOST_STAT)) {
		const struct rule *rule;

		rule = RULE(ruleno);
		if (rule->sconv != 0)
			logmsgx(IPA_LOG_INFO, "rule %s: this rule did not "
			    "receive %s%"PRIu64" units of statistics, this "
			    "statistics is lost", rule->name,
			    rule->sconv_pos ? "" : "-", rule->sconv);
	}
	return (0);
}

static int
ac_deinit(void)
{
	return (0);
}

static int
ac_get_rule_stat(unsigned int stat_generation ATTR_UNUSED, int newstat,
    unsigned int ruleno, int *addition, uint64_t *chunk)
{
	struct rule *rule;

	rule = RULE(ruleno);
	if (newstat) {
		*chunk = 0;
		*addition = 1;
	} else {
		*chunk = rule->sconv;
		*addition = rule->sconv_pos;
	}
	rule->sconv = 0;
	rule->sconv_pos = 1;
	return (0);
}

static int
ac_set_rule_active(unsigned int ruleno, int active)
{
	struct rule *rule;

	rule = RULE(ruleno);
	if (active)
		rule->flags |= RULEF_ACTIVE;
	else
		rule->flags &= ~RULEF_ACTIVE;
	return (0);
}

static int
db_pre_init(void)
{
	logmsgx(IPA_LOG_INFO, "module version " PACKAGE_VERSION);
	return (0);
}

/* ARGSUSED2 */
static int
db_init_statrule(unsigned int ruleno, const char *rule_name,
    const char *rule_info ATTR_UNUSED)
{
	struct rule *rule;

	rule = RULE(ruleno);
	if (rule->dest.rule == NULL) {
		logmsgx(IPA_LOG_ERR, "rule %s: this rule uses module's "
		    "database, but it does not have the conv:dest parameter",
		    rule_name);
		return (-1);
	}
	rule->sorig_pos = 1;
	rule->flags |= RULEF_DB_INITED;
	return (0);
}

/* ARGSUSED */
static int
db_deinit_rule(unsigned int ruleno ATTR_UNUSED)
{
	if (!(quiet & QUIETF_NOT_CONVERTED)) {
		const struct rule *rule;

		rule = RULE(ruleno);
		if (rule->sorig != 0)
			logmsgx(IPA_LOG_INFO, "rule %s: this rule has not "
			    "converted %s%"PRIu64" units of statistics",
			    rule->name, rule->sorig_pos ? "" : "-",
			    rule->sorig);
	}
	return (0);
}

static int
db_deinit(void)
{
	struct plan *plan, *plan_next;
	struct tint *tint, *tint_next;
	unsigned int i;

	for (i = 0; i < nrules; ++i)
		mem_free(RULE(i), m_anon);
	memfunc->marray_deinit(rules_ptr_marray);

	STAILQ_FOREACH_SAFE(plan, &plans_list, link, plan_next) {
		mem_free(plan->name, m_anon);
		mem_free(plan->file, m_anon);
		SLIST_FOREACH_SAFE(tint, &plan->tint_list, link, tint_next)
			mem_free(tint, m_anon);
		mem_free(plan, m_anon);
	}

	return (0);
}

/*
 * Return the most specific time interval in the given plan for
 * the given date.
 */
static void
plan_select_tint(struct plan *plan, const ipa_tm *tm)
{
	struct tint *tint;
	unsigned int sec, sec1_close;

	sec = tm->tm_hour * SECONDS_IN_HOUR + tm->tm_min * SECONDS_IN_MINUTE +
	    tm->tm_sec;
	sec1_close = SECONDS_IN_DAY;
	SLIST_FOREACH(tint, &plan->tint_list, link) {
		switch (tint->type) {
		case TINT_TYPE_YMD:
			if (tint->year != tm->tm_year ||
			    tint->mon != tm->tm_mon ||
			    tint->day != tm->tm_mday)
				continue;
			break;
		case TINT_TYPE_MD:
			if (tint->mon != tm->tm_mon ||
			    tint->day != tm->tm_mday)
				continue;
			break;
		case TINT_TYPE_M:
			if (tint->mon != tm->tm_mon)
				continue;
			break;
		case TINT_TYPE_WDAY:
			if (tint->day != tm->tm_wday)
				continue;
			break;
		case TINT_TYPE_INT:
			break;
		case TINT_TYPE_DEFAULT:
			tint->sec2 = sec1_close;
			goto found;
		}
		if (sec >= tint->sec1) {
			if (sec <= tint->sec2) {
				/* Inside interval. */
				break;
			}
		} else {
			if (sec1_close > tint->sec1) {
				/* Update closest sec1. */
				sec1_close = tint->sec1;
			}
		}
	}
found:
	if (debug) {
		char buf[TINT_TYPE_MAXLEN];

		tint_to_str(tint, buf, sizeof(buf));
		logmsgx(IPA_LOG_INFO, "plan %s: using time interval '%s' "
		    "%"PRIu32" -> %"PRIu32, plan->name, buf, tint->sorig,
		    tint->sconv);
	}
	plan->tint = tint;
	plan->year = tm->tm_year;
	plan->mon = tm->tm_mon;
	plan->mday = tm->tm_mday;
}

static int
db_update_rule(unsigned int ruleno, const uint64_t *cnt_ptr, const ipa_tm *tm)
{
	uint64_t chunk, cnt;
	const struct tint *tint;
	struct rule *rule, *dest;
	unsigned int sec;
	char chunk_pos;

	rule = RULE(ruleno);
	dest = rule->dest.rule;
	cnt = *cnt_ptr;

	/* If the destination rule is inactive, then do not convert. */
	if (!(dest->flags & RULEF_ACTIVE))
		goto done;

	/* If statistics was not changed, then do not convert it. */
	if (rule->cnt == cnt)
		goto done;

	/*
	 * Check if it is the same current time interval.  Since
	 * db_update_rule() is called always for the current day need
	 * to check only hours and minutes.  Also if time goes back,
	 * then ipa calls db_append_rule() and only then db_update_rule().
	 */
	sec = tm->tm_hour * SECONDS_IN_HOUR + tm->tm_min * SECONDS_IN_MINUTE +
	    tm->tm_sec;
	if (sec > rule->plan->tint->sec2)
		plan_select_tint(rule->plan, tm);
	tint = rule->plan->tint;

	/*
	 * If statistics is zeroed, then do not convert it and forget
	 * about not converted original statistics.
	 */
	if (tint->conv == TINT_CONV_ZERO) {
		rule->sorig = 0;
		rule->sorig_pos = 1;
		goto done;
	}

	if (rule->cnt < cnt) {
		/* Statistics increases. */
		chunk = cnt - rule->cnt;
		if (rule->sorig_pos) {
			/* Not converted statistics is positive. */
			if (chunk > UINT64_MAX - rule->sorig) {
				logmsgx(IPA_LOG_ERR, "rule %s: positive chunk "
				    "of original statistics overflowed",
				    rule->name);
				goto failed;
			}
			chunk += rule->sorig;
			chunk_pos = 1;
		} else if (chunk >= rule->sorig) {
			/*
			 * Not converted original statistics is negative,
			 * but current chunk is greater than or equal to it.
			 */
			chunk -= rule->sorig;
			rule->sorig_pos = 1;
			chunk_pos = 1;
		} else {
			/*
			 * Not converted original statistics is negative
			 * and current chunk is less than it.
			 */
			chunk = rule->sorig - chunk;
			chunk_pos = 0;
		}
	} else {
		/* Statistics decreases. */
		chunk = rule->cnt - cnt;
		if (!rule->sorig_pos) {
			/* Not converted statistics is negative. */
			if (chunk > UINT64_MAX - rule->sorig) {
				logmsgx(IPA_LOG_ERR, "rule %s: negative chunk "
				    "of original statistics overflowed",
				    rule->name);
				goto failed;
			}
			chunk += rule->sorig;
			chunk_pos = 0;
		} else if (rule->sorig >= chunk) {
			/*
			 * Not converted original statistics is positive and
			 * it is greater than or equal to the current chunk.
			 */
			chunk = rule->sorig - chunk;
			chunk_pos = 1;
		} else {
			/*
			 * Not converted original statistics is positive,
			 * but current chunk is greater than it.
			 */
			chunk -= rule->sorig;
			chunk_pos = 0;
		}
	}

	/*
	 * Convert statistics and get the rest of not converted original
	 * statistics.  If converted chunk is zero, then do nothing.
	 */
	rule->sorig = chunk % tint->sorig;
	chunk /= tint->sorig;
	if (chunk == 0)
		goto done;
	if (tint->conv == TINT_CONV_INCREASE && chunk > tint->max_div_sconv) {
		logmsgx(IPA_LOG_ERR, "rule %s: %s chunk of converted "
		    "statistics overflowed",
		    chunk_pos ? "positive" : "negative", rule->name);
		goto failed;
	}
	chunk *= tint->sconv;

	if (chunk_pos) {
		/* Chunk is positive. */
		if (dest->sconv_pos) {
			/* Converted statistics is also positive. */
			if (chunk > UINT64_MAX - dest->sconv) {
				logmsgx(IPA_LOG_ERR, "rule %s: internal "
				   "positive statistics of destination rule "
				   "%s overflowed", rule->name, dest->name);
				goto failed;
			}
			dest->sconv += chunk;
		} else if (chunk >= dest->sconv) {
			/*
			 * Converted statistics is negative, but current
			 * chunk is greater than or equal to it.
			 */
			dest->sconv = chunk - dest->sconv;
			dest->sconv_pos = 1;
		} else {
			/*
			 * Converted statistics is negative and current
			 * chunk is less than it.
			 */
			dest->sconv -= chunk;
		}
	} else {
		/* Chunk is negative. */
		if (!dest->sconv_pos) {
			/* Converted statistics is also negative. */
			if (chunk > UINT64_MAX - dest->sconv) {
				logmsgx(IPA_LOG_ERR, "rule %s: internal "
				    "negative statistics of destination rule "
				    "%s overflowed", rule->name, dest->name);
				goto failed;
			}
			dest->sconv += chunk;
		} else if (dest->sconv >= chunk) {
			/*
			 * Converted statistics is positive and it is
			 * greater than or equal to the current chunk.
			 */
			dest->sconv -= chunk;
		} else {
			/*
			 * Converted statistics is positive, but current
			 * chunk is greater than it.
			 */
			dest->sconv = chunk - dest->sconv;
			dest->sconv_pos = 0;
		}
	}

done:
	rule->cnt = cnt;
	return (0);

failed:
	logmsgx(IPA_LOG_ERR, "this means that something is wrong in "
	    "configuration");
	return (-1);
}

static int
db_append_rule(unsigned int ruleno, const uint64_t *cnt_ptr, const ipa_tm *tm)
{
	struct rule *rule;

	rule = RULE(ruleno);

	/*
	 * Forget about previous counter, but do not forget about
	 * not converted statistics.
	 */
	rule->cnt = 0;

	/*
	 * Since db_append_rule() can be called at any time need to
	 * check year, month and month day of the current time interval.
	 * db_update_rule() is called always for the current day.
	 */
	if (rule->plan->year != tm->tm_year ||
	    rule->plan->mon != tm->tm_mon ||
	    rule->plan->mday != tm->tm_mday)
		plan_select_tint(rule->plan, tm);

	return (db_update_rule(ruleno, cnt_ptr, tm));
}

#if IPA_AC_MOD_API_VERSION == 1
static ipa_conf_sect conf_sect_tbl[] = {
	{ NULL, 0, 0, NULL, NULL, IPA_CONF_TYPE_MISC, NULL, NULL }
};

static ipa_conf_param ac_conf_param_tbl[] = {
	{ NULL, 0, NULL, NULL, IPA_CONF_TYPE_MISC,
	  NULL, NULL
	}
};
#else
# define conf_sect_tbl NULL
# define ac_conf_param_tbl NULL
#endif

struct ipa_ac_mod IPA_AC_MOD_ENTRY(ipa_conv) = {
	IPA_AC_MOD_API_VERSION,	/* api_ver			*/
	MOD_FLAGS,		/* mod_flags			*/
	MOD_NAME,		/* ac_name			*/
	NULL,			/* suppfunc			*/
	NULL,			/* memfunc			*/
	MOD_AC_CONF_PREFIX,	/* conf_prefix			*/
	conf_sect_tbl,		/* conf_sect_tbl		*/
	ac_conf_param_tbl,	/* conf_param_tbl		*/
	ac_conf_init,		/* conf_init			*/
	conf_deinit,		/* conf_deinit			*/
	ac_conf_event,		/* conf_event			*/
	ac_conf_mimic_real,	/* conf_mimic_real		*/
	NULL,			/* conf_inherit			*/
	ac_conf_show,		/* conf_show			*/
	ac_pre_init,		/* ac_pre_init			*/
	NULL,			/* ac_init_autorule		*/
	NULL,			/* ac_init_dynrule		*/
	ac_init_statrule,	/* ac_init_statrule		*/
	ac_db_init,		/* ac_init			*/
	ac_deinit_rule,		/* ac_deinit_rule		*/
	NULL,			/* ac_deinit_autorule		*/
	ac_deinit,		/* ac_deinit			*/
	NULL,			/* ac_get_stat			*/
	ac_get_rule_stat,	/* ac_get_rule_stat		*/
	NULL,			/* ac_set_autorule_active	*/
	ac_set_rule_active,	/* ac_set_rule_active		*/
	NULL,			/* ac_set_limit_active		*/
	NULL,			/* ac_set_threshold_active	*/
	NULL,			/* ac_limit_event		*/
	NULL,			/* ac_threshold_event		*/
	NULL,			/* ac_create_rule		*/
	NULL			/* ac_delete_rule		*/
};

static const unsigned int sect_root[] = { IPA_CONF_SECT_ROOT, 0 };
static const unsigned int sect_rule[] = { IPA_CONF_SECT_RULE, 0 };

static ipa_conf_param db_conf_param_tbl[] = {
	{ "load", 2, NULL, NULL, IPA_CONF_TYPE_MISC,
	  sect_root, parse_load
	},
	{ "dest", 1, NULL, NULL, IPA_CONF_TYPE_MISC,
	  sect_rule, parse_dest
	},
	{ "plan", 1, NULL, NULL, IPA_CONF_TYPE_MISC,
	  sect_rule, parse_plan
	},
	{ "debug", 1, NULL, NULL, IPA_CONF_TYPE_UINT32,
	  sect_root, parse_debug
	},
	{ "scale", 1, NULL, NULL, IPA_CONF_TYPE_BOOLEAN,
	  sect_root, parse_scale
	},
	{ "quiet", 1, NULL, NULL, IPA_CONF_TYPE_UINT32,
	  sect_root, parse_quiet
	},
	{ NULL, 0, NULL, NULL, IPA_CONF_TYPE_MISC,
	  NULL, NULL
	}
};

struct ipa_db_mod IPA_DB_MOD_ENTRY(ipa_conv) = {
	IPA_DB_MOD_API_VERSION,	/* api_ver			*/
	MOD_FLAGS,		/* mod_flags			*/
	MOD_NAME,		/* db_name			*/
	NULL,			/* suppfunc			*/
	NULL,			/* memfunc			*/
	MOD_DB_CONF_PREFIX,	/* conf_prefix			*/
	conf_sect_tbl,		/* conf_sect_tbl		*/
	db_conf_param_tbl,	/* conf_param_tbl		*/
	db_conf_init,		/* conf_init			*/
	conf_deinit,		/* conf_deinit			*/
	db_conf_event,		/* conf_event			*/
	db_conf_mimic_real,	/* conf_mimic_real		*/
	NULL,			/* conf_inherit			*/
	db_conf_show,		/* conf_show			*/
	db_pre_init,		/* db_pre_init			*/
	NULL,			/* db_init_dynrule		*/
	db_init_statrule,	/* db_init_statrule		*/
	NULL,			/* db_init_dynlimit		*/
	NULL,			/* db_init_statlimit		*/
	NULL,			/* db_init_dynthreshold		*/
	NULL,			/* db_init_statthreshold	*/
	ac_db_init,		/* db_init			*/
	NULL,			/* db_get_limit_state		*/
	NULL,			/* db_set_limit_state		*/
	NULL,			/* db_get_threshold_state	*/
	NULL,			/* db_set_threshold_state	*/
	NULL,			/* db_deinit_threshold		*/
	NULL,			/* db_deinit_limit		*/
	db_deinit_rule,		/* db_deinit_rule		*/
	db_deinit,		/* db_deinit			*/
	db_append_rule,		/* db_append_rule		*/
	db_update_rule,		/* db_update_rule		*/
	NULL,			/* db_update_limit		*/
	NULL,			/* db_limit_event		*/
	NULL,			/* db_update_threshold		*/
	NULL,			/* db_set_rule_active		*/
	NULL,			/* db_set_limit_active		*/
	NULL			/* db_set_threshold_active	*/
};
